from dataclasses import dataclass, asdict
from decimal import Decimal
from typing import Any
from canatax.enums import ProvinceOrTerritory
from canatax.utils import to_currency


@dataclass
class BaseTaxEstimate:
    
    province: ProvinceOrTerritory

    def _prettify(self, d:dict[str, Any]) -> dict[str,Any]:
        d2 = {}
        for k, v in d.items():
            if isinstance(v, Decimal):
                d2[k] = to_currency(v)
            elif isinstance(v, ProvinceOrTerritory):
                d2[k] = v.value
            else:
                d2[k] = v
        return d2

    def to_dict(self, prettify:bool=False) -> dict[str, Any]:
        """Convert the object's attributes to a dictionary representation.

        Args:
            prettify (bool): If True, applies formatting to the dictionary values.
                            - Floats are formatted as currency strings.
                            - `ProvinceOrTerritory` enum values are converted to their string representation.
                            Defaults to False.

        Returns:
            dict[str, Any]: A dictionary containing the object's attributes. 
                            If `prettify` is True, the dictionary values are formatted accordingly.
        """
        d = asdict(self)
        if prettify:
            d = self._prettify(d)
        return d
    

@dataclass
class SalesTaxEstimate(BaseTaxEstimate):
    
    before_tax: Decimal
    gst: Decimal
    pst: Decimal
    hst: Decimal
    qst: Decimal
    tax_total: Decimal
    after_tax: Decimal


@dataclass
class IncomeTaxEstimate(BaseTaxEstimate):
    
    gross_income: Decimal
    federal_tax: Decimal
    provincial_tax: Decimal
    cpp: Decimal
    ei: Decimal
    qpip: Decimal
    qpp: Decimal
    total_tax: Decimal
    net_income: Decimal


from decimal import Decimal


class BaseSalesTaxRate:

    GST = 0
    PST = 0
    HST = 0
    QST = 0

    @property
    def rate(self) -> Decimal:
        applicable_taxes = [tax for tax in self.tax_types if tax > 0]
        if not applicable_taxes:
            raise AttributeError(f"`{self.__class__.__name__}`: GST, PST, HST, and QST are all 0!")
        return Decimal(sum(applicable_taxes))

    @property
    def tax_types(self) -> list[Decimal]:
        return [Decimal(self.GST), Decimal(self.PST), Decimal(self.HST), Decimal(self.QST)]




from dataclasses import dataclass
from ..base import BaseContribution

# CPP:
# https://www.canada.ca/en/revenue-agency/news/newsroom/tax-tips/tax-tips-2024/canada-revenue-agency-announces-maximum-pensionable-earnings-contributions-2025.html
#
# EI & EIQuebec:
# https://www.canada.ca/en/revenue-agency/services/tax/businesses/topics/payroll/payroll-deductions-contributions/employment-insurance-ei/ei-premium-rates-maximums.html
#
# QPP:
# https://www.revenuquebec.ca/en/businesses/source-deductions-and-employer-contributions/calculating-source-deductions-and-employer-contributions/quebec-pension-plan-contributions/maximum-pensionable-salary-or-wages-and-contribution-rate/
#
# QPIP:
# https://www.quebec.ca/nouvelles/actualites/details/maintien-des-taux-de-cotisation-au-regime-quebecois-dassurance-parentale-en-2025-56294



@dataclass(frozen=True)
class EI(BaseContribution):

    rate = 1.64
    max_earnings = 65700


@dataclass(frozen=True)
class EIQuebec(EI):
    
    rate = 1.31
    
    
@dataclass(frozen=True)
class CPP(BaseContribution):

    rate = 5.95
    max_earnings = 71300
    exemption = 3500

    additional_rate = 4.00
    additional_min = 71300
    additional_max = 81200

    @property
    def additional_rate_decimal(self) -> float:
        return self.additional_rate / 100


@dataclass(frozen=True)
class QPP(CPP):

    rate = 6.40
 
    
@dataclass(frozen=True)
class QPIP(BaseContribution):
    
    max_earnings = 98000
    rate = 0.494



from decimal import Decimal
from canatax.calculators.base_calculator import BaseCalculator
from canatax.enums import ProvinceOrTerritory, TaxType
from canatax.exc import CanataxError
from canatax.rates.income.current_tax import ProvincialIncomeTaxRate
from canatax.tax_estimate import IncomeTaxEstimate
from canatax.rates.income.current_tax import *
from canatax.utils import decimal_round
from canatax.rates.income.current_contributions import Contributions


class IncomeTaxCalculator(BaseCalculator):

    contributions = Contributions()

    def __init__(self, income: int | float | Decimal, province: ProvinceOrTerritory | str):

        income = self._decimalize(income)
        super().__init__(province=province)
        self.income = decimal_round(income)
        self.federal_tax_rate = FederalIncomeTaxRate()
        self.provincial_tax_rate = self._get_tax_rate(TaxType.INCOME)

    def _get_tax_rate(self, tax_type: TaxType) -> ProvincialIncomeTaxRate:
        tax_rate = super()._get_tax_rate(tax_type)
        if isinstance(tax_rate, ProvincialIncomeTaxRate):
            return tax_rate
        raise CanataxError(f"{self.__class__.__name__} failed to load tax rate")

    def _calculate(self) -> IncomeTaxEstimate: 
        federal_tax_rate, provincial_tax_rate = self._tax()
        ei = self._ei()
        if self.is_quebec():
            cpp = Decimal(0)
            qpp = self._cpp()
        else:
            cpp = self._cpp()
            qpp = Decimal(0)
        qpip = self._qpip()
        total_tax = federal_tax_rate + provincial_tax_rate + ei + cpp + qpip
        print(self.income)
        print(total_tax)
        
        net_income = self.income - total_tax
        return IncomeTaxEstimate(
            province=self.province,
            gross_income=self.income,
            federal_tax=federal_tax_rate,
            provincial_tax=provincial_tax_rate,
            ei=ei,
            cpp=cpp,
            qpp=qpp,
            qpip=qpip,
            total_tax=total_tax,
            net_income=net_income,
        )

    @classmethod
    def calculate(cls, income: float | int | Decimal, province: str | ProvinceOrTerritory) -> IncomeTaxEstimate:
        calculator = cls(income=income, province=province)
        return calculator._calculate()

    def _cpp(self) -> Decimal:
        
        if self.is_quebec():
            cpp = self.contributions.qpp
        else:
            cpp = self.contributions.cpp

        base_max = cpp.max_earnings - cpp.exemption
        base_contrib = min(self.income, base_max) * cpp.rate_decimal

        if self.income < cpp.additional_min:
            additional_contrib = Decimal(0)
        else:
            cpp2_income = Decimal(min(self.income, cpp.additional_max) - cpp.additional_min)
            additional_contrib = Decimal(cpp2_income * Decimal(cpp.additional_rate_decimal))

        total = base_contrib + additional_contrib
        return decimal_round(total)

    def _ei(self) -> Decimal:
        if self.is_quebec():
            ei = self.contributions.ei_quebec
        else:
            ei = self.contributions.ei
        ei_float = min(self.income, ei.max_earnings) * ei.rate_decimal
        return decimal_round(ei_float)
        
    def _qpip(self) -> Decimal:
        qpip = self.contributions.qpip
        if not self.is_quebec():
            return Decimal(0)
        
        return decimal_round(min(self.income, qpip.max_earnings) * qpip.rate_decimal)
        
    def _tax(self) -> tuple[Decimal, Decimal]:
        """Return federal tax and provincial tax as a tuple object."""
        federal_tax = decimal_round(self.federal_tax_rate.calculate_tax(self.income - self.federal_tax_rate.get_bpa(self.income)))
        provincial_tax = decimal_round(self.provincial_tax_rate.calculate_tax(self.income - decimal_round(self.provincial_tax_rate.BPA)))
        return federal_tax, provincial_tax
    
        


from decimal import Decimal
import unittest
from canatax.calculators import SalesTaxCalculator
from canatax.enums import ProvinceOrTerritory
from canatax.exc import InvalidProvinceError
from canatax.tax_estimate import SalesTaxEstimate


class TestSalesTaxCalculator(unittest.TestCase):

    # def setUp(self):
    #     self.alberta_calculator = SalesTaxCalculator(province=ProvinceOrTerritory.ALBERTA)
    #     self.bc_calculator = SalesTaxCalculator(province=ProvinceOrTerritory.BRITISH_COLUMBIA)

    def test_invalid_province(self):
        invalid_provinces = ['InvalidProvince', 'ZZ', '123', None, ProvinceOrTerritory]
        for invalid_province in invalid_provinces:
            with self.assertRaises(InvalidProvinceError):
                SalesTaxCalculator(province=invalid_province)


    def test_sales_tax_alberta(self):
        estimate = SalesTaxCalculator.calculate(100.0, "AB")
        self.assertEqual(estimate.gst, Decimal(5.0))
        self.assertEqual(estimate.pst, Decimal(0.0))
        self.assertEqual(estimate.hst, Decimal(0.0))
        self.assertEqual(estimate.tax_total, Decimal(5.0))
        self.assertEqual(estimate.after_tax, Decimal(105.0))


    def test_sales_tax_british_columbia(self):
        estimate = SalesTaxCalculator.calculate(100.0, "BC")
        self.assertEqual(estimate.gst, Decimal(5.0))
        self.assertEqual(estimate.pst, Decimal(7.0))
        self.assertEqual(estimate.hst, Decimal(0.0))
        self.assertEqual(estimate.tax_total, Decimal(12.0))
        self.assertEqual(estimate.tax_total, Decimal(112.0))


if __name__ == '__main__':
    unittest.main()


from canatax.rates.sales.base import BaseSalesTaxRate

# 2025
# https://www.canada.ca/en/revenue-agency/services/tax/businesses/topics/gst-hst-businesses/charge-collect-which-rate/calculator.html
# https://www.revenuquebec.ca/en/businesses/consumption-taxes/gsthst-and-qst/basic-rules-for-applying-the-gsthst-and-qst/tables-of-gst-and-qst-rates/


class AlbertaSalesTaxRate(BaseSalesTaxRate):
    
    GST = 5
    PST = 0
    HST = 0
    QST = 0


class BritishColumbiaSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 7
    HST = 0
    QST = 0


class ManitobaSalesTaxRate(BaseSalesTaxRate):

    GST = 5
    PST = 7
    HST = 0
    QST = 0


class NewBrunswickSalesTaxRate(BaseSalesTaxRate):
        
    GST = 0
    PST = 0
    HST = 15
    QST = 0


class NewfoundlandSalesTaxRate(BaseSalesTaxRate):
        
    GST = 0
    PST = 0
    HST = 15
    QST = 0


class NorthwestTerritoriesSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 0
    HST = 0
    QST = 0


class NovaScotiaSalesTaxRate(BaseSalesTaxRate):
        
    GST = 0
    PST = 0
    HST = 14
    QST = 0


class NunavutSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 0
    HST = 0
    QST = 0


class OntarioSalesTaxRate(BaseSalesTaxRate):
        
    GST = 0
    PST = 0
    HST = 13
    QST = 0


class PEISalesTaxRate(BaseSalesTaxRate):
        
    GST = 0
    PST = 0
    HST = 15
    QST = 0
    

class QuebecSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 0
    HST = 0
    QST = 9.975


class SaskatchewanSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 6
    HST = 0
    QST = 0


class YukonSalesTaxRate(BaseSalesTaxRate):

    GST = 5
    PST = 0
    HST = 0
    QST = 0



from decimal import Decimal, ROUND_HALF_UP
from canatax.calculators.base_calculator import BaseCalculator
from canatax.exc import CanataxError
from canatax.enums import ProvinceOrTerritory, TaxType
from canatax.rates.income.current_tax import ProvincialIncomeTaxRate
from canatax.rates.sales.current_sales_tax import BaseSalesTaxRate
from canatax.tax_estimate import SalesTaxEstimate
from canatax.utils import percent_to_decimal, decimal_round


class SalesTaxCalculator(BaseCalculator):

    def __init__(self, province: ProvinceOrTerritory | str):
        super().__init__(province=province)
        self.tax_rate = self._get_tax_rate(TaxType.SALES)

    def _get_tax_rate(self, tax_type: TaxType) -> BaseSalesTaxRate:
        tax_rate = super()._get_tax_rate(tax_type)
        if isinstance(tax_rate, BaseSalesTaxRate):
            return tax_rate
        raise CanataxError(f"{self.__class__.__name__} failed to load tax rate")

    def _calculate(self, amount: float | int | Decimal) -> SalesTaxEstimate:
        """Calculate sales tax for a given amount based on the tax rates for the initialized province or territory.

        This method calculates the GST, PST, and HST based on the provided amount and the applicable tax rates
        for the current province or territory. It raises an `InvalidDollarAmount` exception if the input amount is invalid.

        Args:
            amount (float): The amount to calculate sales tax on.

        Returns:
            SalesTaxEstimate: An object containing detailed information about the calculated sales tax, including
                            GST, PST, HST, total tax, and the after-tax total.

        Raises:
            InvalidDollarAmount: If the provided amount is negative, `None`, or cannot be converted to a valid decimal.
        """
        amount = self._decimalize(amount)
        gst_total = decimal_round((amount * percent_to_decimal(self.tax_rate.GST))) if self.tax_rate.GST else Decimal('0.00')
        pst_total = decimal_round((amount * percent_to_decimal(self.tax_rate.PST))) if self.tax_rate.PST else Decimal('0.00')
        hst_total = decimal_round((amount * percent_to_decimal(self.tax_rate.HST))) if self.tax_rate.HST else Decimal('0.00')
        qst_total = decimal_round((amount * percent_to_decimal(self.tax_rate.QST))) if self.tax_rate.QST else Decimal('0.00')
        tax_total = decimal_round((gst_total + pst_total + hst_total + qst_total))
        after_tax_total = decimal_round((amount + tax_total))
        return SalesTaxEstimate(
            province=self.province,
            before_tax=amount,
            tax_total=tax_total,
            after_tax=after_tax_total,
            gst=gst_total,
            pst=pst_total,
            hst=hst_total,
            qst=qst_total,
        )

    @classmethod
    def calculate(
            cls, 
            amount: float | int | Decimal, 
            province: str | ProvinceOrTerritory,
    ) -> SalesTaxEstimate:
        calculator = cls(province=province)
        return calculator._calculate(amount)



from abc import ABC
from decimal import Decimal, InvalidOperation
from typing import Type

from canatax.enums import *
from canatax.exc import InvalidProvinceError, InvalidDollarAmount
from canatax.rates.income.current_tax import *
from canatax.rates.sales.current_sales_tax import *



class BaseCalculator(ABC):

    province: ProvinceOrTerritory

    PROVINCE_MAPPING: dict[
        ProvinceOrTerritory, tuple[Type[ProvincialIncomeTaxRate], Type[BaseSalesTaxRate]]
    ] = {
        ProvinceOrTerritory.ALBERTA : (AlbertaIncomeTaxRate, AlbertaSalesTaxRate),
        ProvinceOrTerritory.BRITISH_COLUMBIA: (BritishColumbiaIncomeTaxRate, BritishColumbiaSalesTaxRate),
        ProvinceOrTerritory.MANITOBA : (ManitobaIncomeTaxRate, ManitobaSalesTaxRate), 
        ProvinceOrTerritory.ONTARIO : (OntarioIncomeTaxRate, OntarioSalesTaxRate),
        ProvinceOrTerritory.NEW_BRUNSWICK : (NewBrunswickIncomeTaxRate, NewBrunswickSalesTaxRate),
        ProvinceOrTerritory.NEWFOUNDLAND : (NewfoundlandIncomeTaxRate, NewfoundlandSalesTaxRate),
        ProvinceOrTerritory.NORTHWEST_TERRITORIES : (NorthwestTerritoriesIncomeTaxRate, NorthwestTerritoriesSalesTaxRate),
        ProvinceOrTerritory.NOVA_SCOTIA : (NovaScotiaIncomeTaxRate, NovaScotiaSalesTaxRate),
        ProvinceOrTerritory.NUNAVUT : (NunavutIncomeTaxRate, NunavutSalesTaxRate),
        ProvinceOrTerritory.PRINCE_EDWARD_ISLAND: (PEIIncomeTaxRate, PEISalesTaxRate),
        ProvinceOrTerritory.QUEBEC: (QuebecIncomeTaxRate, QuebecSalesTaxRate),
        ProvinceOrTerritory.SASKATCHEWAN : (SaskatchewanIncomeTaxRate, SaskatchewanSalesTaxRate),
        ProvinceOrTerritory.YUKON : (YukonIncomeTaxRate, YukonSalesTaxRate),
    }

    def __init__(self, province: str | ProvinceOrTerritory):
        """Initializes the calculator with a province or territory.

        Args:
            province (str | ProvinceOrTerritory): The province or territory as a string or an enum.

        Raises:
            InvalidProvinceError: If the province or territory is not valid.
        """
        self.province = self._coerce_province(province)
        
        
    def _coerce_province(self, province: str | ProvinceOrTerritory) -> ProvinceOrTerritory:
        if not isinstance(province, ProvinceOrTerritory):
            try:
                return ProvinceOrTerritory(province.upper())
            except ValueError as e:
                raise InvalidProvinceError(province) from e

        return province

    def _decimalize(self, amount:int|float|Decimal) -> Decimal:
        try:
            decimal_amount = Decimal(amount)
        except (ValueError, TypeError, InvalidOperation) as e:
            raise InvalidDollarAmount(amount) from e
        else:
            if decimal_amount.is_infinite() or decimal_amount.is_nan():
                raise InvalidDollarAmount(amount)
            if decimal_amount < 0:
                raise InvalidDollarAmount(amount)
        return decimal_amount

    def is_quebec(self) -> bool:
        match self.province:
            case ProvinceOrTerritory.QUEBEC:
                return True
            case _:
                return False

    def _get_tax_rate(self, tax_type:TaxType) -> ProvincialIncomeTaxRate | BaseSalesTaxRate:
        tax_rate_tuple = self.PROVINCE_MAPPING[self.province]
        match tax_type:
            case TaxType.INCOME:
                return tax_rate_tuple[0]()
            case TaxType.SALES:
                return tax_rate_tuple[1]()
            case _:
                raise ValueError(f"Invalid param tax_type: `{tax_type}` ")

#  (AlbertaIncomeTaxRate, AlbertaSalesTaxRate),


from canatax.rates.sales.base import BaseSalesTaxRate


class AlbertaSalesTaxRate(BaseSalesTaxRate):
    
    GST = 5
    PST = None
    HST = None


class BritishColumbiaSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 7
    HST = None


class ManitobaSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 7
    HST = None


class QuebecSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 9.975
    HST = None


class PEISalesTaxRate(BaseSalesTaxRate):
        
    GST = None
    PST = None
    HST = 15


class NovaScotiaSalesTaxRate(BaseSalesTaxRate):
        
    GST = None
    PST = None
    HST = 15


class NewBrunswickSalesTaxRate(BaseSalesTaxRate):
        
    GST = None
    PST = None
    HST = 15


class NewfoundlandSalesTaxRate(BaseSalesTaxRate):
        
    GST = None
    PST = None
    HST = 15


class NunavutSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = None
    HST = None


class NorthwestTerritoriesSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = None
    HST = None


class SaskatchewanSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = 6
    HST = None


class OntarioSalesTaxRate(BaseSalesTaxRate):
        
    GST = None
    PST = None
    HST = 13


class YukonSalesTaxRate(BaseSalesTaxRate):
        
    GST = 5
    PST = None
    HST = None



from decimal import Decimal, ROUND_HALF_UP


def to_currency(n:float|Decimal) -> str:
    return f"{n:,.2f}"


def percent_to_decimal(n:float|Decimal) -> Decimal:
    return Decimal(n / 100)


def decimal_round(n: float | int | Decimal):
    return Decimal(n).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)



from setuptools import setup, find_packages

setup(
    name='canatax',
    version='2.0.1',
    description='An easy-to-use, dependency-free Canadian sales and income tax calculator.',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Michael Pearce',
    author_email='firstflush@protonmail.com',
    url='https://github.com/firstflush/canatax',
    license='MIT',
    packages=find_packages(exclude=["tests*", "*.tests"]),
    include_package_data=True,
    install_requires=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Topic :: Office/Business :: Financial :: Accounting',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    keywords='canada tax cpp ei qpp gst hst pst qst income sales payroll',
    project_urls={
        'Homepage': 'https://github.com/firstflush/canatax',
        'Bug Tracker': 'https://github.com/firstflush/canatax/issues',
        'Buy Me a Coffee': 'https://www.buymeacoffee.com/FirstFlush',
    },
)



from decimal import Decimal
from canatax.rates.income.base import BaseIncomeTaxRate, ProvincialIncomeTaxRate

# 2025 Provincial and federal tax rates:
# ----
# https://www.canada.ca/en/revenue-agency/services/tax/individuals/frequently-asked-questions-individuals/canadian-income-tax-rates-individuals-current-previous-years.html
# https://www.revenuquebec.ca/en/citizens/income-tax-return/completing-your-income-tax-return/income-tax-rates/


class FederalIncomeTaxRate(BaseIncomeTaxRate):
    """
    15% 	on the portion of taxable income that is $57,375 or less, plus
    20.5% 	on the portion of taxable income over $57,375 up to $114,750, plus
    26% 	on the portion of taxable income over $114,750 up to $177,882, plus
    29% 	on the portion of taxable income over $177,882 up to $253,414, plus
    33% 	on the portion of taxable income over $253,414
    """
    
    _BPA_MIN = Decimal(14538) # https://www.canada.ca/en/revenue-agency/services/forms-publications/payroll/t4032-payroll-deductions-tables/t4032on-jan/t4032on-january-general-information.html
    _BPA_MAX = Decimal(15705)
    _BPA_PHASE_OUT_START = Decimal("173205")
    _BPA_PHASE_OUT_END = Decimal("246752")
    
    brackets = [
        (15, 57375),
        (20.5, 114750),
        (26, 177882),
        (29, 253414),
        (33, float('inf')),
    ]

    @classmethod
    def get_bpa(cls, income: Decimal) -> Decimal:
        if income <= cls._BPA_PHASE_OUT_START:
            bpa = cls._BPA_MAX
        elif income >= cls._BPA_PHASE_OUT_END:
            bpa = cls._BPA_MIN 
        else:
            reduction_ratio = ((income - cls._BPA_PHASE_OUT_START) / (cls._BPA_PHASE_OUT_END - cls._BPA_PHASE_OUT_START))
            bpa = cls._BPA_MAX - (reduction_ratio * (cls._BPA_MAX - cls._BPA_MIN))
        return bpa


class AlbertaIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    10% 	on the portion of taxable income that is $151,234 or less, plus
    12% 	on the portion of taxable income over $151,234 up to $181,481, plus
    13% 	on the portion of taxable income over $181,481 up to $241,974, plus
    14% 	on the portion of taxable income over $241,974 up to $362,961, plus
    15% 	on the portion of taxable income over $362,961
    """
    
    BPA = 22323  # https://www.atb.com/wealth/good-advice/tax/alberta-provincial-budget-2025/

    brackets = [
        (10, 151234),
        (12, 181481),
        (13, 241974),
        (14, 362961),
        (15, float('inf')),
    ]


class BritishColumbiaIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    5.06% 	on the portion of taxable income that is $49,279 or less, plus
    7.7% 	on the portion of taxable income over $49,279 up to $98,560, plus
    10.5% 	on the portion of taxable income over $98,560 up to $113,158, plus
    12.29% 	on the portion of taxable income over $113,158 up to $137,407, plus
    14.7% 	on the portion of taxable income over $137,407 up to $186,306, plus
    16.8% 	on the portion of taxable income over $186,306 up to $259,829, plus
    20.5% 	on the portion of taxable income over $259,829
    """
    
    BPA = 12932  # https://www2.gov.bc.ca/gov/content/taxes/income-taxes/personal/credits/basic
    
    brackets = [
        (5.06, 49279),
        (7.7, 98560),
        (10.5, 113158),
        (12.29, 137407),
        (14.7, 186306),
        (16.8, 259829),
        (20.5, float('inf')),
    ]


class ManitobaIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    10.8% 	on the portion of taxable income that is $47,564 or less, plus
    12.75% 	on the portion of taxable income over $47,564 up to $101,200, plus
    17.4% 	on the portion of taxable income over $101,200
    """
    
    BPA = 15780  # https://www.pwc.com/ca/en/services/tax/budgets/2025/manitoba.html
    
    brackets = [
        (10.8, 47564),
        (12.75, 101200),
        (17.4, float('inf')),
    ]


class NewBrunswickIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    9.4% 	on the portion of taxable income that is $51,306 or less, plus
    14% 	on the portion of taxable income over $51,306 up to $102,614, plus
    16% 	on the portion of taxable income over $102,614 up to $190,060, plus
    19.5% 	on the portion of taxable income over $190,060    
    """
    
    BPA = 13396 # https://www2.gnb.ca/content/gnb/en/departments/finance/taxes/personal.html
    
    brackets = [
        (9.4, 51306),
        (14, 102614),
        (16, 190060),
        (19.5, float('inf')),
    ]


class NewfoundlandIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    8.7% 	on the portion of taxable income that is $44,192 or less, plus
    14.5% 	on the portion of taxable income over $44,192 up to $88,382, plus
    15.8% 	on the portion of taxable income over $88,382 up to $157,792, plus
    17.8% 	on the portion of taxable income over $157,792 up to $220,910, plus
    19.8% 	on the portion of taxable income over $220,910 up to $282,214, plus
    20.8% 	on the portion of taxable income over $282,214 up to $564,429, plus
    21.3% 	on the portion of taxable income over $564,429 up to $1,128,858, plus
    21.8% 	on the portion of taxable income over $1,128,858
    """
    
    BPA = 10808 # https://turbotax.intuit.ca/tips/newfoundland-and-labrador-provincial-taxes-and-credits-569
    
    brackets = [
        (8.7, 44192),
        (14.5, 88382),
        (15.8, 157792),
        (17.8, 220910),
        (19.8, 282214),
        (20.8, 564429),
        (21.3, 1128858),
        (21.8, float('inf')),
    ]


class NorthwestTerritoriesIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    5.9% 	on the portion of taxable income that is $51,964 or less, plus
    8.6% 	on the portion of taxable income over $51,964 up to $103,930, plus
    12.2% 	on the portion of taxable income over $103,930 up to $168,967, plus
    14.05% 	on the portion of taxable income over $168,967
    """
    
    BPA = 15705 # https://turbotax.intuit.ca/tips/northwest-territories-tax-rates-and-the-most-popular-credits-deductions-programs-and-rebates-5073
    
    brackets = [
        (5.9, 51964),
        (8.6, 103930),
        (12.2, 168967),
        (14.05, float('inf')),
    ]


class NovaScotiaIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    8.79% 	on the portion of taxable income that is $30,507 or less, plus
    14.95% 	on the portion of taxable income over $30,507 up to $61,015, plus
    16.67% 	on the portion of taxable income over $61,015 up to $95,883, plus
    17.5% 	on the portion of taxable income over $95,883 up to $154,650, plus
    21% 	on the portion of taxable income over $154,650
    """
    
    BPA = 14744 # https://www.canada.ca/en/revenue-agency/services/forms-publications/payroll/t4032-payroll-deductions-tables/t4032ns-july/t4032ns-july-general-information.html#_Toc337712806
    
    brackets = [
        (8.79, 30507),
        (14.95, 61015),
        (16.67, 95883),
        (17.5, 154650),
        (21, float('inf')),
    ]


class NunavutIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    4% 	on the portion of taxable income that is $54,707 or less, plus
    7% 	on the portion of taxable income over $54,707 up to $109,413, plus
    9% 	on the portion of taxable income over $109,413 up to $177,881, plus
    11.5% 	on the portion of taxable income over $177,881    
    """
    
    BPA = 18767 # https://turbotax.intuit.ca/tips/nunavut-territorial-taxes-and-credits-5071
    
    brackets = [
        (4, 54707),
        (7, 109413),
        (9, 177881),
        (11.5, float('inf')),
    ]


class OntarioIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    5.05% 	on the portion of taxable income that is $52,886 or less, plus
    9.15% 	on the portion of taxable income over $52,886 up to $105,775, plus
    11.16% 	on the portion of taxable income over $105,775 up to $150,000, plus
    12.16% 	on the portion of taxable income over $150,000 up to $220,000, plus
    13.16% 	on the portion of taxable income over $220,000
    """
    
    BPA = 12399 # https://turbotax.intuit.ca/tips/ontario-provincial-taxes-and-credits-574
    
    brackets = [
        (5.05, 52886),
        (9.15, 105775),
        (11.16, 150000),
        (12.16, 220000),
        (13.16, float('inf')),
    ]


class PEIIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    9.5% 	on the portion of taxable income that is $33,328 or less, plus
    13.47% 	on the portion of taxable income over $33,328 up to $64,656, plus
    16.6% 	on the portion of taxable income over $64,656 up to $105,000, plus
    17.62% 	on the portion of taxable income over $105,000 up to $140,000, plus
    19% 	on the portion of taxable income over $140,000    
    """
    
    BPA = 15050 # https://www.canada.ca/en/revenue-agency/services/forms-publications/payroll/t4032-payroll-deductions-tables/t4032pe-july/t4032pe-july-general-information.html#_Toc337712806
    
    brackets = [
        (9.5, 33328),
        (13.47, 64656),
        (16.6, 105000),
        (17.62, 140000),
        (19, float('inf')),
    ]


class QuebecIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    $53,255 or less	14%
    More than $53,255 but not more than $106,495	19%
    More than $106,495 but not more than $129,590	24%
    More than $129,590	25.75%
    """
    
    BPA = 18056 # https://www.revenuquebec.ca/en/citizens/income-tax-return/completing-your-income-tax-return/how-to-complete-your-income-tax-return/line-by-line-help/350-to-398-1-non-refundable-tax-credits/line-350/
    
    brackets = [
        (14, 53255),
        (19, 106495),
        (24, 129590),
        (25.75, float('inf')),
    ]


class SaskatchewanIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    10.5% 	on the portion of taxable income that is $53,463 or less, plus
    12.5% 	on the portion of taxable income over $53,463 up to $152,750, plus
    14.5% 	on the portion of taxable income over $152,750    
    """
    
    BPA = 18491 # https://turbotax.intuit.ca/tips/saskatchewan-provincial-taxes-and-credits-5062
    
    brackets = [
        (10.5, 53463),
        (12.5, 152750),
        (14.5, float('inf')),
    ]


class YukonIncomeTaxRate(ProvincialIncomeTaxRate):
    """ 
    6.4% 	on the portion of taxable income that is $57,375 or less, plus
    9% 	on the portion of taxable income over $57,375 up to $114,750, plus
    10.9% 	on the portion of taxable income over $114,750 up to $177,882, plus
    12.8% 	on the portion of taxable income over $177,882 up to $500,000, plus
    15% 	on the portion of taxable income over $500,000
    """
    
    BPA = 15705 # Yes, same as the fed. Not a typo. https://turbotax.intuit.ca/tips/yukon-territorial-taxes-and-credits-5066
    
    brackets = [
        (6.4, 57375),
        (9, 114750),
        (10.9, 177882),
        (12.8, 500000),
        (15, float('inf')),
    ]


from typing import Any
from canatax.enums import ProvinceOrTerritory

class CanataxError(Exception):
    """Base class to all Canatax related errors."""
    pass


class InvalidProvinceError(CanataxError):
    """Raised when ProvinceOrTerritory enum is passed an invalid value."""
    def __init__(self, province:str|ProvinceOrTerritory):
        message = f"Invalid province or territory: `{province}`. Value must be two-letter string: AB,BC,MB,NB,NL,NS,NT,NU,ON,PE,QC,SK,YK."
        super().__init__(message)


class InvalidDollarAmount(CanataxError):
    """Raised when an invalid amount is passed to a tax calculator."""
    def __init__(self, amount:Any):
        message = f"Invalid dollar amount `{amount}`"
        super().__init__(message)


from canatax.calculators.income_calculator import IncomeTaxCalculator
from canatax.calculators.sales_calculator import SalesTaxCalculator
from canatax.calculators.base_calculator import BaseCalculator


from abc import ABC, abstractmethod
from decimal import Decimal, ROUND_HALF_UP
from canatax.utils import percent_to_decimal

class BaseIncomeTaxRate(ABC):
    
    brackets: list[tuple[float|int, int | float]]

    def calculate_tax(self, income: Decimal) -> Decimal:
        """Returns the tax owed (estimate) on a given income amount."""
        tax_owed = Decimal(0)
        previous_threshold = Decimal(0)
        for rate, threshold in self.brackets:
            if income > threshold:
                tax_owed += (Decimal(threshold) - Decimal(previous_threshold)) * percent_to_decimal(rate)
                previous_threshold = threshold
            else:
                tax_owed += (income - Decimal(previous_threshold)) * percent_to_decimal(rate)
                break
        return tax_owed.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)

    @classmethod
    def get_bpa(cls, income: Decimal) -> Decimal:
        if hasattr(cls, "BPA") and not hasattr(cls, "_BPA_MIN") and not hasattr(cls, "_BPA_MAX"):
            return getattr(cls, "BPA")
        elif hasattr(cls, "_BPA_MIN") and hasattr(cls, "_BPA_MAX"):
            raise NotImplementedError(f"{cls.__name__} has a BPA (basic personal amount) range. It must implement its own calculation for BPA.")
        else:
            raise NotImplementedError(f"{cls.__name__} missing BPA (basic personal amount) information.")


class ProvincialIncomeTaxRate(BaseIncomeTaxRate):
    ...


class BaseContribution:

    rate = 0
    max_earnings = 0
    
    @property
    def rate_decimal(self) -> Decimal:
        return Decimal(self.rate / 100)
